//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package auth

import (
	"context"
	"encoding/json"
	"net/http"
	"time"

	oidc "github.com/coreos/go-oidc"
	"golang.org/x/oauth2"
)

type OIDCSession struct {
	State string
	Nonce string
	token *oauth2.Token
}

func NewOIDCSession() (s *OIDCSession, err error) {
	s = &OIDCSession{}
	if s.State, err = generateRandomString(32); err != nil {
		return
	}
	if s.Nonce, err = generateRandomString(32); err != nil {
		return
	}
	return
}

// TODO: check concurrent usage of provider/verifier etc...

type OIDCBackend struct {
	provider     *oidc.Provider
	verifier     *oidc.IDTokenVerifier
	oauth2Config oauth2.Config
}

func NewOIDCBackend(cfg *OIDCConfig) (b *OIDCBackend, err error) {
	// TODO: make ctx a parameter?
	ctx := context.Background()
	b = &OIDCBackend{}
	if b.provider, err = oidc.NewProvider(ctx, cfg.IssuerURL); err != nil {
		return
	}

	oidcConfig := &oidc.Config{
		ClientID: cfg.ClientID,
	}
	b.verifier = b.provider.Verifier(oidcConfig)

	b.oauth2Config = oauth2.Config{
		ClientID:     cfg.ClientID,
		ClientSecret: cfg.ClientSecret,
		Endpoint:     b.provider.Endpoint(),
		RedirectURL:  cfg.CallbackURL,
		Scopes:       []string{oidc.ScopeOpenID, "username", "aura_shows"},
	}

	return
}

// TODO: implement session refresh go-routine
//       see: https://github.com/golang/oauth2/issues/84

func (b *OIDCBackend) CallbackHandler() http.Handler {
	return &oidcCallbackHandler{backend: b}
}

func (b *OIDCBackend) HandleLogin(w http.ResponseWriter, r *http.Request) {
	sid, s := getSessionFromCookie(r)
	if s != nil {
		msg := ""
		if s.Username != "" {
			msg = "You are still logged in, please logout first."
		} else if s.oidc == nil {
			msg = "This is not an OIDC session."
		} else {
			msg = "OIDC login already in progress, retry later."
		}
		http.Error(w, msg, http.StatusConflict)
		return
	}

	os, err := NewOIDCSession()
	if err != nil {
		http.Error(w, "Failed to generate new OIDC session: "+err.Error(), http.StatusInternalServerError)
		return
	}
	s = &Session{oidc: os, Expires: time.Now().Add(time.Minute)}
	if sid, err = auth.sessions.insert(s); err != nil {
		http.Error(w, "Failed to generate new session: "+err.Error(), http.StatusInternalServerError)
		return
	}
	setSessionCookie(w, sid)
	http.Redirect(w, r, b.oauth2Config.AuthCodeURL(s.oidc.State, oidc.Nonce(s.oidc.Nonce)), http.StatusFound)
}

type oidcCallbackHandler struct {
	backend *OIDCBackend
}

func (h *oidcCallbackHandler) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	sid, s := getSessionFromCookie(r)
	if s == nil {
		invalidateSessionCookie(w)
		http.Error(w, "Request does not contain a valid session cookie or session is already expired.", http.StatusUnauthorized)
		return
	}

	if s.oidc == nil {
		http.Error(w, "This is not an OIDC session.", http.StatusConflict)
		return
	}
	if s.Username != "" {
		http.Error(w, "This session is already logged in.", http.StatusConflict)
		return
	}
	if r.URL.Query().Get("state") != s.oidc.State {
		http.Error(w, "OIDC verification failed: state did not match", http.StatusBadRequest)
		return
	}

	oauth2Token, err := h.backend.oauth2Config.Exchange(r.Context(), r.URL.Query().Get("code"))
	if err != nil {
		invalidateSessionCookie(w)
		auth.sessions.remove(sid)
		http.Error(w, "OAuth2 token exchange failed: "+err.Error(), http.StatusBadRequest)
		return
	}
	rawIDToken, ok := oauth2Token.Extra("id_token").(string)
	if !ok {
		invalidateSessionCookie(w)
		auth.sessions.remove(sid)
		http.Error(w, "OIDC verification failed: no id_token field in oauth2 token.", http.StatusInternalServerError)
		return
	}

	// Verify the ID Token signature and nonce.
	idToken, err := h.backend.verifier.Verify(r.Context(), rawIDToken)
	if err != nil {
		invalidateSessionCookie(w)
		auth.sessions.remove(sid)
		http.Error(w, "OAuth2 ID token verification failed: "+err.Error(), http.StatusInternalServerError)
		return
	}
	if idToken.Nonce != s.oidc.Nonce {
		invalidateSessionCookie(w)
		auth.sessions.remove(sid)
		http.Error(w, "OAuth2 ID token  verification failed: invalid nonce", http.StatusInternalServerError)
		return
	}

	userInfo, err := h.backend.provider.UserInfo(r.Context(), oauth2.StaticTokenSource(oauth2Token))
	if err != nil {
		invalidateSessionCookie(w)
		auth.sessions.remove(sid)
		http.Error(w, "Fetching OIDC UserInfo failed: "+err.Error(), http.StatusInternalServerError)
		return
	}

	newS := &Session{}
	if err := userInfo.Claims(newS); err != nil {
		http.Error(w, err.Error(), http.StatusInternalServerError)
		return
	}
	maxAge := defaultAge
	if auth.config.Sessions.MaxAge > 0 {
		maxAge = auth.config.Sessions.MaxAge
	}
	newS.Expires = time.Now().Add(maxAge * time.Second)
	newS.oidc = &OIDCSession{State: s.oidc.State, Nonce: s.oidc.Nonce}
	newS.oidc.token = oauth2Token
	if err = auth.sessions.update(sid, newS); err != nil {
		invalidateSessionCookie(w)
		auth.sessions.remove(sid)
		http.Error(w, "Updating session failed: "+err.Error(), http.StatusInternalServerError)
		return
	}
	setSessionCookie(w, sid) // reset the cookie to update max-age

	data, _ := json.MarshalIndent(newS, "", "    ")
	w.Write(data)
}
