//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package auth

import (
	"context"
	"errors"
	"math"
	"net/http"
	"sync"
	"time"
)

const (
	defaultAge        = 24 * time.Hour
	SessionCookieName = "aura-tank-session"
)

type Session struct {
	Expires  time.Time `json:"expires"`
	Username string    `json:"username"`
	IsAdmin  bool      `json:"is-admin"`
	Shows    []string  `json:"shows"`
	oidc     *OIDCSession
}

var (
	anonAllowNone = &Session{Username: "anonymous", IsAdmin: false, Shows: []string{}}
	anonAllowAll  = &Session{Username: "anonymous", IsAdmin: true, Shows: []string{}}
)

func (s *Session) Expired() bool {
	return s.Expires.Before(time.Now())
}

func getSessionFromCookie(r *http.Request) (string, *Session) {
	sc, err := r.Cookie(SessionCookieName)
	if sc == nil && err != nil {
		return "", nil
	}

	sid := sc.Value
	s := auth.sessions.get(sid)
	if s == nil {
		return "", nil
	}
	if s.Expired() {
		auth.sessions.remove(sid)
		return "", nil
	}

	return sid, s
}

func newSessionCookie(sid string, maxAge int) *http.Cookie {
	sc := &http.Cookie{Name: SessionCookieName}
	sc.Value = sid
	sc.HttpOnly = true
	sc.MaxAge = maxAge

	sc.Secure = auth.config.Sessions.Cookies.Secure
	sc.Path = auth.config.Sessions.Cookies.Path
	if sc.Path == "" {
		sc.Path = "/"
	}
	sc.Domain = auth.config.Sessions.Cookies.Domain
	sc.SameSite = http.SameSite(auth.config.Sessions.Cookies.SameSite)
	return sc
}

func setSessionCookie(w http.ResponseWriter, sid string) {
	maxAge := int(math.Ceil(defaultAge.Seconds()))
	if auth.config.Sessions.MaxAge > 0 {
		maxAge = int(math.Ceil(auth.config.Sessions.MaxAge.Seconds()))
	}
	http.SetCookie(w, newSessionCookie(sid, maxAge))
}

func invalidateSessionCookie(w http.ResponseWriter) {
	http.SetCookie(w, newSessionCookie("invalid", -1))
}

func attachSessionToRequest(r *http.Request, s *Session) *http.Request {
	ctx := context.WithValue(r.Context(), sessionContextKey, s)
	return r.WithContext(ctx)
}

func SessionFromRequest(r *http.Request) (*Session, bool) {
	s, ok := r.Context().Value(sessionContextKey).(*Session)
	return s, ok
}

type SessionManager struct {
	mutex    sync.RWMutex
	sessions map[string]*Session
}

func NewSessionManager() (sm *SessionManager, err error) {
	sm = &SessionManager{}
	sm.sessions = make(map[string]*Session)
	go sm.runMaintenance()
	return
}

func (sm *SessionManager) runMaintenance() {
	t := time.NewTicker(time.Minute * 10)
	for {
		<-t.C
		sm.cleanupExpired()
	}
}

func (sm *SessionManager) insert(s *Session) (id string, err error) {
	if id, err = generateRandomString(32); err != nil {
		return
	}

	sm.mutex.Lock()
	defer sm.mutex.Unlock()
	sm.sessions[id] = s
	return
}

func (sm *SessionManager) get(id string) *Session {
	sm.mutex.RLock()
	defer sm.mutex.RUnlock()

	s, ok := sm.sessions[id]
	if !ok {
		return nil
	}
	return s
}

func (sm *SessionManager) update(id string, s *Session) error {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	if _, ok := sm.sessions[id]; !ok {
		return errors.New("session not found.")
	}
	sm.sessions[id] = s
	return nil
}

func (sm *SessionManager) remove(id string) {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	delete(sm.sessions, id)
}

func (sm *SessionManager) cleanupExpired() {
	sm.mutex.Lock()
	defer sm.mutex.Unlock()

	for id, s := range sm.sessions {
		if s.Expired() {
			delete(sm.sessions, id)
		}
	}
}
